from enum import Enum, auto
from types import DynamicClassAttribute
from pandas.core.frame import DataFrame


class EnumBase(Enum):
    def __str__(self):
        return self.name.lower()

    @DynamicClassAttribute
    def name(self):
        """The name of the Enum member."""
        return self._name_.lower()


class Status(EnumBase):
    SUCCESS = auto()
    FAIL = auto()
    SKIP = auto()
    PROCESSING = auto()


class DupPolicy(EnumBase):
    REPLACE = auto()
    SKIP = auto()
    RAISE_ERROR = auto()


def generate_insert_sql(
    source_df: DataFrame,
    tablename: str,
    chunksize: int = None,
    unique_key: list[str] = None,
    skip_on_conflict: bool = False,
) -> str | list[str]:
    """
    Generate SQL INSERT or UPDATE statements for a given DataFrame.

    Args:
        source_df (DataFrame): The source DataFrame containing the data to be inserted into the database.
        tablename (str): The name of the database table where the data will be inserted.
        chunksize (int, optional): The size of each chunk of data to be inserted. If specified, the function generates
            multiple SQL statements, each containing a chunk of rows. If None, a single SQL statement is generated for
            the entire DataFrame. Default is None.
        unique_key (list of str, optional): The column(s) to be used as a unique key to check for duplicates in the
            database table. If duplicates are found, the function generates UPDATE statements instead of INSERT
            statements. Default is None.
        skip_on_conflict (bool, optional): If True, generates SQL statements to skip insertion on conflict (DO NOTHING).
            If False, generates UPDATE statements on conflict. Default is False.

    Returns:
        str | list[str]: If chunksize is provided, a list of SQL statements is returned, with each statement
            containing a chunk of data. If chunksize is None, a single SQL statement is returned as a string,
            containing all the data from the DataFrame.

    Examples:
    ```
    # Generate SQL INSERT or UPDATE statements for the entire DataFrame
    sql_statements = generate_insert_sql(df, 'your_table_name', unique_key=['column1'])

    # Generate SQL INSERT or SKIP statements for the entire DataFrame
    sql_statements = generate_insert_sql(df, 'your_table_name', unique_key=['column1'], skip_on_conflict=True)

    # Generate SQL INSERT or UPDATE statements in chunks of 2 rows
    sql_statements = generate_insert_sql(df, 'your_table_name', chunksize=2, unique_key=['column1', 'column2'])
    ```
    """
    if chunksize:
        sql_text_list: list[str] = []

        for i in range(0, len(source_df), chunksize):
            chunk_df = source_df.iloc[i : i + chunksize]
            if unique_key:
                insert_values = ", ".join(
                    [str(tuple(row)) for _, row in chunk_df.iterrows()]
                )
                if skip_on_conflict:
                    sql_text_list.append(
                        f"INSERT INTO {tablename} ({', '.join(source_df.columns)}) "
                        f"VALUES {insert_values} "
                        f"ON CONFLICT ({', '.join(unique_key)}) DO NOTHING"
                    )
                else:
                    update_values = ", ".join(
                        [f"{col} = EXCLUDED.{col}" for col in source_df.columns]
                    )
                    sql_text_list.append(
                        f"INSERT INTO {tablename} ({', '.join(source_df.columns)}) "
                        f"VALUES {insert_values} "
                        f"ON CONFLICT ({', '.join(unique_key)}) DO UPDATE SET {update_values}"
                    )
            else:
                insert_values = ", ".join(
                    [str(tuple(row)) for _, row in chunk_df.iterrows()]
                )
                sql_text_list.append(
                    f"INSERT INTO {tablename} ({', '.join(source_df.columns)}) VALUES {insert_values}"
                )

        return sql_text_list
    else:
        sql_text: list[str] = []
        for _, row in source_df.iterrows():
            if unique_key:
                if skip_on_conflict:
                    sql_text.append(
                        f"INSERT INTO {tablename} ({', '.join(source_df.columns)}) "
                        f"VALUES {str(tuple(row.values))} "
                        f"ON CONFLICT ({', '.join(unique_key)}) DO NOTHING"
                    )
                else:
                    update_values = ", ".join(
                        [f"{col} = EXCLUDED.{col}" for col in source_df.columns]
                    )
                    sql_text.append(
                        f"INSERT INTO {tablename} ({', '.join(source_df.columns)}) "
                        f"VALUES {str(tuple(row.values))} "
                        f"ON CONFLICT ({', '.join(unique_key)}) DO UPDATE SET {update_values}"
                    )
            else:
                sql_text.append(
                    f"INSERT INTO {tablename} ({', '.join(source_df.columns)}) "
                    f"VALUES {str(tuple(row.values))}"
                )

        return sql_text



import os 
from lichens.db.models import EtlProgMng
from lichens.db.utils import add_etl as _add_etl
from sqlalchemy import Engine, create_engine
from typing import Any

def add_etl(name:str, src_folder:os.PathLike, dst_folder:os.PathLike, json_setting:dict, update_by:int, con:str | Engine)->str:
    """_summary_

    Args:
        name (str): Name of the program
        src_folder (os.PathLike): source folder
        dst_folder (os.PathLike): archive folder
        json_setting (dict): setting for program
        update_by (int): user id.
        con (str | Engine): connection strings or an sqlalchemy.Engine

    Returns:
        str: _description_
    """
    if isinstance(con, str):
        con = create_engine(con)

    orm_dict:dict = {
        "name":name,
        "src_folder":src_folder,
        "dst_folder":dst_folder,
        "json_setting":json_setting,
        "last_log":None,
        "update_by": update_by
    }
    etl_name:str = _add_etl(orm=EtlProgMng(**orm_dict), con=con)
    return etl_name

etl_prog_mng_template:dict[str, Any] = {
        "name": 'program_name',
        "src_folder": 'path/to/files',
        "dst_folder":'path/to/archive',
        "json_setting": {},
        "last_log": {},
        "update_by": 0
    }


from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session

# postgresql+psycopg2://scott:tiger@host/dbname
DB_URL = ''
engine = create_engine(DB_URL, echo=True)
SessionLocal:Session = sessionmaker(autocommit=False, autoflush=False, bind=engine)

def get_db_session():
    """Create a user auth db session.

    Returns:
        Session: a sqlalchemy Session instance used to conduct db operations. 
    """
    db = SessionLocal()
    try:
        return db
    except Exception:
        db.rollback()
        raise
    finally:
        db.close()





from os import PathLike
import os
from time import sleep
from typing import Literal, Callable
from crontab import CronTab
import pendulum
import shutil
from sqlalchemy import Engine, create_engine, text
from sqlalchemy.orm import Session, sessionmaker
from lichens.db.models import EtlProcHist, EtlProgMng
from lichens.errors.db_errors import *
from lichens.utils import Status, generate_insert_sql, DupPolicy
from pandas.core.frame import DataFrame
import abc
from logging import getLogger

log = getLogger()


class EtlManager:
    def __init__(
        self,
        constr: str,
        name: str,
    ) -> None:
        """An ETL manager coworks with Pharmquer

        Args:
            constr (str): database connection string.
            name (str): name of ETL.
        """
        self.constr: str = constr
        self.name: str = name
        self._engine: Engine = None
        self._etl_setting: EtlProgMng = None
        self.src_folder: PathLike = None
        self.dst_folder: dict[str, PathLike] = {}
        self.conf:dict = None

        self._fetch_config()

    def _fetch_config(
        self,
    ) -> None:
        try:
            self._engine: Engine = create_engine(self.constr)
            with Session(self._engine) as sess:
                self._etl_setting = (
                    sess.query(EtlProgMng).filter(EtlProgMng.name == self.name).first()
                )
            if not self._etl_setting:
                raise ProgramNotFoundError(f"Name={self.name} not found. You can use lichens.tools.add_etl() to add one first.")
                
            self.src_folder = self._etl_setting.src_folder
            self.dst_folder[Status.FAIL.name] = os.path.join(
                self._etl_setting.dst_folder, Status.FAIL.name
            )
            self.dst_folder[Status.SUCCESS.name] = os.path.join(
                self._etl_setting.dst_folder, Status.SUCCESS.name
            )
            self.dst_folder[Status.SKIP.name] = os.path.join(
                self._etl_setting.dst_folder, Status.SKIP.name
            )
            self.conf:dict = self._etl_setting.json_setting

        except ProgramNotFoundError as e:
            raise e
        except Exception as e:
            raise DatabaseConnectingFailed(e)
        
    def reload_conf(self):
        self._fetch_config()

    def move(self, src:os.PathLike, status: Literal[Status.FAIL, Status.SUCCESS, Status.SKIP]) -> None:
        """Move the processed file to the destination folder according to the status.

        Args:
            src (PathLike): The path of source file. 
            status (Literal[&#39;fail&#39;, &#39;skip&#39;, &#39;success&#39;]): The process status
        """
        fn:os.PathLike = os.path.basename(src)
        if not os.path.exists(self.dst_folder.get(status)):
            os.makedirs(self.dst_folder.get(status))
        dst:os.PathLike = os.path.join(self.dst_folder.get(status), fn)
        try: 
            shutil.move(src, dst)
        except Exception as e:
            raise e

    def update_status(
        self,
        filename: str,
        user_id:int,
        status: Literal[Status.FAIL, Status.SUCCESS, Status.SKIP, Status.PROCESSING],
        last_log: dict[str, str] = None,
    ) -> None:
        """update the current status and log to the database.
        Args:
            filename (str): processed filename.
            user_id (int): the user who upload or process the file. 
            status (Literal[&#39;fail&#39;, &#39;skip&#39;, &#39;success&#39;, &#39;processing&#39;]): The current status.
            last_log (dict[str, str]): log in json. Recommended&Default={ "status": "processing", "filename":"sample.csv", "update_dtt": pendulum.now()}.
        """
        if not last_log:
            last_log = {
                "status": status,
                "filename": filename,
                "update_dtt": pendulum.now(),
            }
        with Session(self._engine) as s:
            try:
                print()
                s.query(EtlProgMng).filter(
                    EtlProgMng.id == self._etl_setting.id
                ).update({EtlProgMng.last_log: last_log})

                updated_rows:int = s.query(EtlProcHist)\
                    .filter(EtlProcHist.file_name == filename)\
                    .update({
                        EtlProcHist.status: status,
                        EtlProcHist.last_log: last_log,
                    })
                if updated_rows==0:
                    new:dict = {
                        "file_name":filename,
                        "etl_id":self._etl_setting.id,
                        "status": status,
                        "update_by": user_id,
                        "last_log": last_log,
                        "create_dtt": pendulum.now().__str__(),
                    }
                    s.add(EtlProcHist(**new))
                s.commit()
            except Exception as e:
                s.rollback()
                raise UpdateStatusFailed(f"""File: {filename}. Errors: {e}""")

    def load_df(
        self,
        df: DataFrame,
        tablename: str,
        schema:str=None,
        if_exists: Literal[
            DupPolicy.REPLACE, DupPolicy.RAISE_ERROR, DupPolicy.SKIP
        ] = DupPolicy.RAISE_ERROR.name,
        chunksize:int=None,
        unique_key:list[str]=None,
    )->None:
        """Load a DataFrame to the target table. 

        Args:
            df (DataFrame): the transformed pd.DataFrame, which has identical schema to the target table. 
            tablename (str): targer table name. If the schema is NOT default, then is MUST BE ADDED.
            schema (str): schema name
            if_exists (Literal[ &#39;replace&#39;, , &#39;skip&#39;, , &#39;raise_error&#39;, ], optional): . Defaults to "replace".
        """
        try:
            sess:Session = Session(self._engine)
        except Exception as e:
            raise e
        
        if if_exists!=DupPolicy.SKIP.name and not unique_key:
            raise UniqueKeyMissedError(f"Please specify the unique key from {str(tuple(df.columns))}")
        
        if schema:
            tablename = f"{schema}.{tablename}"

        def _do_insert(unique_key_:list[str]=None, skip_on_conflict_:bool=False):
            _sql: list[str] | str = generate_insert_sql(df, tablename, chunksize, unique_key_, skip_on_conflict_)
            _sql = _sql if isinstance(_sql, list) else [_sql, ]
            return list(map(sess.execute, [text(a) for a in _sql]))
        
        try:
            if if_exists == DupPolicy.REPLACE.name:
                _ = _do_insert(unique_key, True)
            elif if_exists == DupPolicy.SKIP.name:
                _ = _do_insert(unique_key, False)
            else: #  if_exists == DupPolicy.RAISE_ERROR.name
                _ = _do_insert(None, False)
            sess.commit()
        except Exception as e:
            sess.rollback()
            raise InsertInterruptedError(e)
        finally:
            sess.close()

    def run_as_schtask(self, func:Callable, crontab:str, times_:int=-1, *args, **kwargs)->None:
        """
        Run a function based on a cron-like schedule using a Schtasks approach.

        Args:
            func (Callable): The function to be executed.
            crontab (str): A string representing the cron-like schedule for function execution.
            times_ (int, optional): The number of times the function should be executed. 
                If set to -1 (default), the function runs indefinitely based on the cron schedule.

        Returns:
        None

        Example:
        ```
        # Create an instance of EtlManager
        em = EtlManager()

        # Define a function to be executed
        def your_function():
            # Your function logic here
            pass

        # Run the function every minute for a total of 5 times
        em.run_as_schtask(your_function, '*/1 * * * *', times_=5)
        ```

        Note:
        - The `crontab` parameter follows the standard cron format.
        - If `times_` is set to -1, the function runs indefinitely based on the cron schedule.
        - The `func` parameter should be a callable function with no arguments.

        """
        runs:int = 0
        while runs<times_:
            sec_waiting_next_run:int = CronTab(crontab).next()
            func(*args, **kwargs)
            runs += 1
            log.info(f'>>> Wait {sec_waiting_next_run} seconds for next run <<<')
            if runs<times_:
                sleep(sec_waiting_next_run)

    def scheduled(self, crontab:str, times_:int=-1)->None:
        """
        Decorator to schedule a function to run based on a cron-like schedule.

        Args:
            crontab (str): A string representing the cron-like schedule for function execution.
            times_ (int, optional): The number of times the function should be executed.
                If set to -1 (default), the function runs indefinitely based on the cron schedule.

        Returns:
        Callable: The decorated function.

        Example:
        ```
        # Create an instance of EtlManager
        em = EtlManager()

        # Decorate the function with the scheduled decorator
        @em.scheduled(crontab='*/1 * * * *', times_=5)
        def your_scheduled_function():
            # Your function logic here
            pass
        ```

        Note:
        - The `crontab` parameter follows the standard cron format.
        - If `times_` is set to -1, the function runs indefinitely based on the cron schedule.
        - The decorated function should not have any arguments.

        """
        def decorator(func: Callable) -> Callable:
            def wrapper(*args, **kwargs):
                self.run_as_schtask(func, crontab, times_, *args, **kwargs)
            return wrapper
        return decorator
                


        


import logging
import string

logger = logging.getLogger()
logger.setLevel(logging.INFO)

class CustomFormatter(logging.Formatter):

    grey = "\x1b[38;20m"
    green = "\x1b[32;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format:string.Template = string.Template("%(asctime)s - %(name)s - $color%(levelname)s\x1b[0m - (%(filename)s:%(lineno)d) - %(message)s ")

    FORMATS = {
        logging.DEBUG: format.substitute({"color":grey}),
        logging.INFO: format.substitute({"color":green}),
        logging.WARNING: format.substitute({"color":yellow}),
        logging.ERROR:format.substitute({"color":red}),
        logging.CRITICAL: format.substitute({"color":bold_red})
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)
    
# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)


from sqlalchemy import Column, ForeignKey, Integer, MetaData, String, func, DateTime
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import declarative_base

Base = declarative_base(metadata=MetaData(schema='pharmquer'))


class EtlProcHist(Base):
    __tablename__ = "etl_proc_hist"
    
    id = Column(Integer, nullable=False, primary_key=True)
    file_name  = Column(String(256), nullable=False, )
    etl_id = Column(Integer, ForeignKey('etl_prog_mng.id'))
    status = Column(String(16), nullable=False,)
    update_by = Column(Integer)
    last_log = Column(JSONB)
    create_dtt = Column(DateTime, default=func.now(), nullable=False)
    update_dtt = Column(DateTime, default=func.now(), nullable=False)


class EtlProgMng(Base):
    __tablename__ = "etl_prog_mng"

    id = Column(Integer, nullable=False, primary_key=True)
    name  = Column(String(128), nullable=False, unique=True)
    src_folder = Column(String(256), nullable=False,)
    dst_folder = Column(String(256), nullable=False,)
    json_setting = Column(JSONB, nullable=False,)
    last_log = Column(JSONB)
    update_by = Column(Integer)
    update_dtt = Column(DateTime, default=func.now(), nullable=False)


class ExceptionBase(Exception):
    def __init__(self, msg: str | None = None, *args: object) -> None:
        super().__init__(msg, *args)
        self.msg = msg
    def __repr__(self) -> str:
        return super().__repr__()
    

class DatabaseConnectingFailed(ExceptionBase):
    def __repr__(self) -> str:
        return f'Failed to connect to Database. Detail: {self.msg}'
    

class UpdateStatusFailed(ExceptionBase):
    def __repr__(self) -> str:
        return f'Failed to update the status. Detail: {self.msg}'
    

class RowExistsAlreadyError(ExceptionBase):
    def __repr__(self) -> str:
        return f'File has duplicated row. Detail={self.msg}'
    

class UniqueKeyMissedError(ExceptionBase):
    def __repr__(self) -> str:
        return super().__repr__()
    

class InsertInterruptedError(ExceptionBase):
    def __repr__(self) -> str:
        return super().__repr__()
    
class ProgramNotFoundError(ExceptionBase):
    def __repr__(self) -> str:
        return super().__repr__()


import json
import os, click
import sys

from sqlalchemy import create_engine
from lichens.db.models import EtlProgMng
from lichens.db.utils import add_etl as _add_etl
from lichens.tools import etl_prog_mng_template

os.environ['LICHENS_HOME'] = os.path.dirname(__file__)
os.environ['ALEMBIC_CONFIG'] = os.path.join(os.environ.get('LICHENS_HOME'), "./db/migrations/alembic.ini")

_orm_script_dir:os.PathLike = os.path.join(os.environ['LICHENS_HOME'], 'db/migrations/versions')
if not os.path.exists(_orm_script_dir):
    os.mkdir(_orm_script_dir)

@click.group()
def cli():
    pass


@click.command(help="Make and migrate the system tables.")
@click.option(
    '-c',
    '--connection-string',
    type=str,
    help="Set the string in format: driver://user:pass@localhost:port/dbname. ",
    show_default=True
)
def migrate(connection_string:str):
    try:
        os.environ['LICHENS_SQLALCHEMY_URL'] = connection_string
        _ = os.popen("""alembic revision --autogenerate -m \"init\"""").read()
        _ = os.popen("""alembic upgrade head""").read()
    except Exception as e:
        print(e)
        sys.exit(1)

@click.command(help="Add an ETL program setting from a JSON")
@click.option(
    '-c',
    '--connection-string',
    type=str,
    help="Set the string in format: driver://user:pass@localhost:port/dbname. ",
    show_default=True
)
@click.option(
    '-j',
    '--json-config',
    type=str,
    help="the ETL program info in json. You can use `get_template` to generate an blank one.",
    show_default=True
)
def add_etl(config:str, connection_string:str):
    try:
        with open(config, 'r') as f:
            orm_conf:dict = json.loads(f.read())
        _ = _add_etl(orm=EtlProgMng(**orm_conf), con=create_engine(connection_string))
        print(f'{orm_conf["name"]} created. ')
    except Exception as e:
        print(e)
        sys.exit(1)

@click.command(help="Generate an ETL program setting to a JSON")
@click.option(
    '-d',
    '--dir',
    type=str,
    default='.',
    help="The directory to generate the json.",
    show_default=True
)
def get_template(dir:os.PathLike)->None:
    with open(os.path.join(dir, 'etl-setting.json'), 'w') as f:
        json.dump(etl_prog_mng_template, f, indent=4)


cli.add_command(migrate)
cli.add_command(add_etl)
cli.add_command(get_template)
cli()


from lichens.db.models import EtlProgMng
from sqlalchemy import Engine
from sqlalchemy.orm import Session

def add_etl(orm:EtlProgMng, con:Engine)->str | None:    
    sess:Session = None
    try:
        sess = Session(con)
        sess.add(orm)
        sess.commit()
        return orm.name
    except Exception as e:
        sess.rollback()
        raise e
    finally:
        if sess:
            sess.close()
